require 'rubygems'
require "optiflag"
require 'plist'

module CommandLineArgs extend OptiFlagSet
  flag "baseDir"
  flag "sourceDir"
  flag "testDir"
  optional_switch_flag "yaml"
  optional_switch_flag "unit_test"
  optional_switch_flag "sanity_check"
  
  and_process!
end

def files_matching(dir, matcher)
  files = []
  Dir.chdir(dir) do
    files = Dir.glob(matcher)
  end
  files.map! { |file| File.join(dir, file) }
  return files
end

base_dir = ARGV.flags.baseDir
base_dir += "/" if base_dir[-1..-1] != "/"
source_dir = ARGV.flags.sourceDir
test_dir = ARGV.flags.testDir

def test_case_start_regexp
  if ARGV.flags.unit_test
    return /def (test_[a-zA-Z0-9_]+)/
  else
    return /it "(.*)"/
  end
end

def rcov_invocation_string(source_dir, test_file, line_nr_and_method_name)
  if ARGV.flags.unit_test
    "rcov --text-counts -x '.*' -i '^#{source_dir}' #{test_file} -- --name #{line_nr_and_method_name[:method_name]}"# 2>/dev/null"
  else
    "rcov --text-counts -x '.*' -i '^#{source_dir}' bin/spec -- -e \"#{line_nr_and_method_name[:method_name]}\" #{test_file}"# 2>/dev/null"
  end
end

if ARGV.flags.sanity_check
  if ARGV.flags.unit_test
    test_files = ["test/block_test.rb"]
    source_files = ["lib/liquid/block.rb"]
  else
    test_files = ["spec/spec/mocks/mock_ordering_spec.rb"]
    source_files = ["lib/spec/mocks/message_expectation.rb"]
  end
else
  if ARGV.flags.unit_test
    test_files = files_matching(test_dir, "**/*_test.rb")
  else
    test_files = files_matching(test_dir, "**/*_spec.rb")
  end
  source_files = files_matching(source_dir, "**/*.rb")
end

tests = {
  :metadata => {
    :base_dir => base_dir,
    :source_dir => source_dir,
    :test_dir => test_dir
  },
  :data => {} # Hashed by "filename#methodname", which returns an array of hashes type {:testfilename, :linenr, :hits} with all test cases with more than 0 hits
}

# for each source file, for each line save to which method it belongs
file_methods_to_range = {}
source_files.each do |source_file|
  file_methods_to_range[source_file] = {}
  current_method = nil
  current_method_start = nil
  file_lines = File.readlines(source_file)
  file_lines.each_with_index do |line, line_nr|
    if line =~ /(^def | def )(.*)$/
      unless current_method.nil?
        file_methods_to_range[source_file][current_method] = current_method_start..(line_nr - 1)
      end
      current_method = $2
      current_method_start = line_nr
    end
  end
  unless current_method.nil?
    file_methods_to_range[source_file][current_method] = current_method_start..(file_lines.size - 1)
  end
end

# For each test file, get an array of line numbers where a test case begins
tests_in_files = {}
test_files.each do |test_file|
  test_lines = File.readlines(test_file)
  line_nrs_and_method_names = []
  test_lines.each_with_index do |line, index|
    if line =~ test_case_start_regexp
      line_nrs_and_method_names ||= []
      line_nrs_and_method_names << {:line_nr => (index + 1), :method_name => $1}
    end
  end
  tests_in_files[test_file] = line_nrs_and_method_names
end

# For each source file, run every single test case watched with rcov.
# The output will give as the number of times each line in the source file was called.
# From the command line it would be rcov --text-counts  -x '.*' -i lib/spec/mocks/message_expectation.rb bin/spec -- spec/spec/mocks/mock_ordering_spec.rb:24
rcov_output_file_sep = "================================================================================\n"
progress_counter = 1
tests_in_files.each do |test_file, line_nrs_and_method_names|
  $stderr.write("Processing #{test_file} (#{progress_counter}/#{tests_in_files.size})\n")
  progress_counter += 1
  line_nrs_and_method_names.each do |line_nr_and_method_name|
    $stderr.write("\t Executing test '#{line_nr_and_method_name[:method_name]}' at line #{line_nr_and_method_name[:line_nr]}\n")
    rcov_result = `#{rcov_invocation_string(source_dir, test_file, line_nr_and_method_name)}`
    rcov_result = rcov_result.split(rcov_output_file_sep)
    if rcov_result.size > 1
      # Extract the report for each source_file
      rcov_result_per_source_file = {}
      for result_line_nr in 0..(rcov_result.size - 1) do
        if source_files.include?(rcov_result[result_line_nr][0..-2]) and !source_files.include?(rcov_result[result_line_nr + 1][0..-2])
          rcov_result_for_file = rcov_result[result_line_nr + 1].split("\n")
          rcov_result_for_file.map! { |result_line| result_line[72..-1].to_i } # Extract the hit count for each line

          rcov_result_per_source_file[rcov_result[result_line_nr][0..-2]] = rcov_result_for_file
          next
        end
      end
      
      rcov_result_per_source_file.each_key do |source_file|
        file_methods_to_range[source_file] and file_methods_to_range[source_file].each do |method_name, range|
          if rcov_result_per_source_file[source_file][range].max > 1
            tests[:data]["#{source_file}"] ||= {}
            tests[:data]["#{source_file}"]["#{method_name}"] ||= []
            tests[:data]["#{source_file}"]["#{method_name}"] << {
              :test_file => test_file,
              :line_nr => line_nr_and_method_name[:line_nr],
              :test_name => line_nr_and_method_name[:method_name],
              :hits => rcov_result_per_source_file[source_file][range].max
            }
          end
        end
      end
    end
  end
end

if ARGV.flags.yaml
  puts tests.to_yaml
else
  puts tests.to_plist
end
